@@ -666,6 +666,8 @@ DEFUN (show_ipv6_ospf6_spf_tree,
   struct ospf6_route *route;
   struct prefix prefix;
 
+  OSPF6_CMD_CHECK_RUNNING ();
+
   ospf6_linkstate_prefix (ospf6->router_id, htonl (0), &prefix);
 
   for (ALL_LIST_ELEMENTS_RO (ospf6->area_list, node, oa))
@@ -701,6 +703,8 @@ DEFUN (show_ipv6_ospf6_area_spf_tree,
   struct ospf6_route *route;
   struct prefix prefix;
 
+  OSPF6_CMD_CHECK_RUNNING ();
+
   ospf6_linkstate_prefix (ospf6->router_id, htonl (0), &prefix);
 
   if (inet_pton (AF_INET, argv[0], &area_id) != 1)
@@ -747,6 +751,8 @@ DEFUN (show_ipv6_ospf6_simulate_spf_tree_root,
   struct ospf6_route_table *spf_table;
   unsigned char tmp_debug_ospf6_spf = 0;
 
+  OSPF6_CMD_CHECK_RUNNING ();
+
   inet_pton (AF_INET, argv[0], &router_id);
   ospf6_linkstate_prefix (router_id, htonl (0), &prefix);
 
@@ -793,10 +799,6 @@ ospf6_area_init (void)
   install_element (VIEW_NODE, &show_ipv6_ospf6_area_spf_tree_cmd);
   install_element (VIEW_NODE, &show_ipv6_ospf6_simulate_spf_tree_root_cmd);
 
-  install_element (ENABLE_NODE, &show_ipv6_ospf6_spf_tree_cmd);
-  install_element (ENABLE_NODE, &show_ipv6_ospf6_area_spf_tree_cmd);
-  install_element (ENABLE_NODE, &show_ipv6_ospf6_simulate_spf_tree_root_cmd);
-
   install_element (OSPF6_NODE, &area_range_cmd);
   install_element (OSPF6_NODE, &area_range_advertise_cmd);
   install_element (OSPF6_NODE, &no_area_range_cmd);
