@@ -234,9 +234,9 @@ ospf6_router_lsa_originate (struct thread *thread)
 
       /* Multiple Router-LSA instance according to size limit setting */
       if ( (oa->router_lsa_size_limit != 0)
-          && ((caddr_t) lsdesc + sizeof (struct ospf6_router_lsdesc) -
-	      /* XXX warning: comparison between signed and unsigned */
-              (caddr_t) buffer > oa->router_lsa_size_limit))
+          && ((size_t)((char *)lsdesc - buffer)
+                 + sizeof (struct ospf6_router_lsdesc)
+               > oa->router_lsa_size_limit))
         {
           if ((caddr_t) lsdesc == (caddr_t) router_lsa +
                                   sizeof (struct ospf6_router_lsa))
@@ -246,30 +246,6 @@ ospf6_router_lsa_originate (struct thread *thread)
               return 0;
             }
 
-          /* Fill LSA Header */
-          lsa_header->age = 0;
-          lsa_header->type = htons (OSPF6_LSTYPE_ROUTER);
-          lsa_header->id = htonl (link_state_id);
-          lsa_header->adv_router = oa->ospf6->router_id;
-          lsa_header->seqnum =
-            ospf6_new_ls_seqnum (lsa_header->type, lsa_header->id,
-                                 lsa_header->adv_router, oa->lsdb);
-          lsa_header->length = htons ((caddr_t) lsdesc - (caddr_t) buffer);
-
-          /* LSA checksum */
-          ospf6_lsa_checksum (lsa_header);
-
-          /* create LSA */
-          lsa = ospf6_lsa_create (lsa_header);
-
-          /* Originate */
-          ospf6_lsa_originate_area (lsa, oa);
-
-          /* Reset setting for consecutive origination */
-          memset ((caddr_t) router_lsa + sizeof (struct ospf6_router_lsa),
-                  0, (caddr_t) lsdesc - (caddr_t) router_lsa);
-          lsdesc = (struct ospf6_router_lsdesc *)
-            ((caddr_t) router_lsa + sizeof (struct ospf6_router_lsa));
           link_state_id ++;
         }
 
@@ -338,7 +314,7 @@ ospf6_router_lsa_originate (struct thread *thread)
   lsa_header->adv_router = oa->ospf6->router_id;
   lsa_header->seqnum =
     ospf6_new_ls_seqnum (lsa_header->type, lsa_header->id,
-                         lsa_header->adv_router, oa->lsdb);
+			 lsa_header->adv_router, oa->lsdb);
   lsa_header->length = htons ((caddr_t) lsdesc - (caddr_t) buffer);
 
   /* LSA checksum */
@@ -1275,14 +1251,13 @@ ospf6_intra_prefix_lsa_add (struct ospf6_lsa *lsa)
         break;
 
       /* Appendix A.4.1.1 */
-      if (CHECK_FLAG(op->prefix_options, OSPF6_PREFIX_OPTION_NU) ||
-	  CHECK_FLAG(op->prefix_options, OSPF6_PREFIX_OPTION_LA))
+      if (CHECK_FLAG(op->prefix_options, OSPF6_PREFIX_OPTION_NU))
 	{
 	  if (IS_OSPF6_DEBUG_EXAMIN (INTRA_PREFIX))
 	    {
 	      ospf6_linkstate_prefix2str ((struct prefix *)OSPF6_PREFIX_BODY(op),
 					  buf, sizeof (buf));
-	      zlog_debug ("%s: Skipping Prefix %s has NU/LA option set",
+	      zlog_debug ("%s: Skipping Prefix %s has NU option set",
 			  __func__, buf);
 	    }
 	  continue;
@@ -1314,8 +1289,8 @@ ospf6_intra_prefix_lsa_add (struct ospf6_lsa *lsa)
         }
       else
         {
-          for (i = 0; ospf6_nexthop_is_set (&ls_entry->nexthop[i]) &&
-               i < OSPF6_MULTI_PATH_LIMIT; i++)
+          for (i = 0; i < OSPF6_MULTI_PATH_LIMIT &&
+               ospf6_nexthop_is_set (&ls_entry->nexthop[i]); i++)
             ospf6_nexthop_copy (&route->nexthop[i], &ls_entry->nexthop[i]);
         }
 
@@ -1339,7 +1314,7 @@ ospf6_intra_prefix_lsa_remove (struct ospf6_lsa *lsa)
   struct ospf6_area *oa;
   struct ospf6_intra_prefix_lsa *intra_prefix_lsa;
   struct prefix prefix;
-  struct ospf6_route *route;
+  struct ospf6_route *route, *nroute;
   int prefix_num;
   struct ospf6_prefix *op;
   char *start, *current, *end;
@@ -1377,8 +1352,9 @@ ospf6_intra_prefix_lsa_remove (struct ospf6_lsa *lsa)
 
       for (ospf6_route_lock (route);
            route && ospf6_route_is_prefix (&prefix, route);
-           route = ospf6_route_next (route))
+           route = nroute)
         {
+          nroute = ospf6_route_next (route);
           if (route->type != OSPF6_DEST_TYPE_NETWORK)
             continue;
           if (route->path.area_id != oa->area_id)
@@ -1408,7 +1384,7 @@ ospf6_intra_prefix_lsa_remove (struct ospf6_lsa *lsa)
 void
 ospf6_intra_route_calculation (struct ospf6_area *oa)
 {
-  struct ospf6_route *route;
+  struct ospf6_route *route, *nroute;
   u_int16_t type;
   struct ospf6_lsa *lsa;
   void (*hook_add) (struct ospf6_route *) = NULL;
@@ -1435,8 +1411,9 @@ ospf6_intra_route_calculation (struct ospf6_area *oa)
   oa->route_table->hook_remove = hook_remove;
 
   for (route = ospf6_route_head (oa->route_table); route;
-       route = ospf6_route_next (route))
+       route = nroute)
     {
+      nroute = ospf6_route_next (route);
       if (CHECK_FLAG (route->flag, OSPF6_ROUTE_REMOVE) &&
           CHECK_FLAG (route->flag, OSPF6_ROUTE_ADD))
         {
@@ -1495,7 +1472,8 @@ ospf6_brouter_debug_print (struct ospf6_route *brouter)
 
   zlog_info ("Brouter: %s via area %s", brouter_name, area_name);
   zlog_info ("  memory: prev: %p this: %p next: %p parent rnode: %p",
-             brouter->prev, brouter, brouter->next, brouter->rnode);
+             (void *)brouter->prev, (void *)brouter, (void *)brouter->next,
+             (void *)brouter->rnode);
   zlog_info ("  type: %d prefix: %s installed: %s changed: %s",
              brouter->type, destination, installed, changed);
   zlog_info ("  lock: %d flags: %s%s%s%s", brouter->lock,
@@ -1515,7 +1493,7 @@ ospf6_brouter_debug_print (struct ospf6_route *brouter)
 void
 ospf6_intra_brouter_calculation (struct ospf6_area *oa)
 {
-  struct ospf6_route *brouter, *copy;
+  struct ospf6_route *brouter, *nbrouter, *copy;
   void (*hook_add) (struct ospf6_route *) = NULL;
   void (*hook_remove) (struct ospf6_route *) = NULL;
   u_int32_t brouter_id;
@@ -1543,7 +1521,7 @@ ospf6_intra_brouter_calculation (struct ospf6_area *oa)
           IS_OSPF6_DEBUG_ROUTE (MEMORY))
         {
           zlog_info ("%p: mark as removing: area %s brouter %s",
-                     brouter, oa->name, brouter_name);
+                     (void *)brouter, oa->name, brouter_name);
           ospf6_brouter_debug_print (brouter);
         }
     }
@@ -1575,7 +1553,7 @@ ospf6_intra_brouter_calculation (struct ospf6_area *oa)
           IS_OSPF6_DEBUG_ROUTE (MEMORY))
         {
           zlog_info ("%p: transfer: area %s brouter %s",
-                     brouter, oa->name, brouter_name);
+                     (void *)brouter, oa->name, brouter_name);
           ospf6_brouter_debug_print (brouter);
         }
     }
@@ -1584,8 +1562,9 @@ ospf6_intra_brouter_calculation (struct ospf6_area *oa)
   oa->ospf6->brouter_table->hook_remove = hook_remove;
 
   for (brouter = ospf6_route_head (oa->ospf6->brouter_table); brouter;
-       brouter = ospf6_route_next (brouter))
+       brouter = nbrouter)
     {
+      nbrouter = ospf6_route_next (brouter);
       brouter_id = ADV_ROUTER_IN_PREFIX (&brouter->prefix);
       inet_ntop (AF_INET, &brouter_id, brouter_name, sizeof (brouter_name));
       
