@@ -47,7 +47,8 @@ static void group_timer_off(struct igmp_group *group);
 static struct igmp_group *find_group_by_addr(struct igmp_sock *igmp,
 					     struct in_addr group_addr);
 
-static int igmp_sock_open(struct in_addr ifaddr, int ifindex, uint32_t pim_options)
+static int igmp_sock_open(struct in_addr ifaddr, ifindex_t ifindex, 
+                          uint32_t pim_options)
 {
   int fd;
   int join = 0;
@@ -110,7 +111,9 @@ static int igmp_sock_open(struct in_addr ifaddr, int ifindex, uint32_t pim_optio
 static void igmp_sock_dump(array_t *igmp_sock_array)
 {
   int size = array_size(igmp_sock_array);
-  for (int i = 0; i < size; ++i) {
+  int i;
+  
+  for (i = 0; i < size; ++i) {
     
     struct igmp_sock *igmp = array_get(igmp_sock_array, i);
     
@@ -952,7 +955,7 @@ static int pim_igmp_read(struct thread *t)
   socklen_t tolen = sizeof(to);
   uint8_t buf[PIM_IGMP_BUFSIZE_READ];
   int len;
-  int ifindex = -1;
+  ifindex_t ifindex = -1;
   int result = -1; /* defaults to bad */
 
   zassert(t);
